/*
	$Id: Auditor.c 232 2006-06-03 18:46:16Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <setjmp.h>
#include <dlfcn.h>
#include <err.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <tf/TestFunnel.h>
#include "StackFrame.h"
#include "Auditor.h"

static struct {
    jmp_buf environment;
    const char *className;
    const char *methodName;
    int timeout;
} the;

void
tf_impl_printStackTrace(void)
{
    StackFrame_printStackTrace(STACKFRAME_CURRENT_FP, 20);
}

void __attribute__((noreturn))
tf_impl_fail(const char *file, int line, const char *s)
{
    fprintf(stderr, "%s:%d: %s#%s() failed: %s\n",
	    file, line, the.className, the.methodName, s);
    StackFrame_printStackTrace(STACKFRAME_CURRENT_FP, 20);
    longjmp(the.environment, 1);
}

static void * __attribute__((noreturn))
run(void *arg UNUSED)
{
    void *h;
    void (*f)(void);
    struct itimerval timeout = {{0, 0}, {0, 0}};

    timeout.it_value.tv_sec = the.timeout;
    setitimer(ITIMER_REAL, &timeout, NULL);
    if ((h = dlopen(the.className, RTLD_NOW)) == NULL) {
	errx(2, "%s: %s", the.className, dlerror());
    }
    if (setjmp(the.environment)) {
	exit(1);
    }
    if ((f = dlsym(h, "tf_initialize")) != NULL) {
	f();
    }
    if ((f = dlsym(h, the.methodName)) == NULL) {
	errx(2, "%s#%s: dlsym() failed.", the.className, the.methodName);
    }
    f();
    if ((f = dlsym(h, "tf_finalize")) != NULL) {
	f();
    }
    (void)dlclose(h);
    exit(0);
}

static void __attribute__((noreturn))
audit(const char *className, const char *methodName)
{
    pthread_t thread;

    the.className = className;
    the.methodName = methodName;
    pthread_create(&thread, NULL, run, NULL);
    pthread_join(thread, NULL);
    abort();
}

void
Auditor_setTimeout(int sec)
{
    the.timeout = sec;
}

int
Auditor_invoke(const char *className, const char *methodName)
{
    pid_t pid;
    int status, e;

    fflush(NULL);
    if ((pid = fork()) < 0) {
	err(1, "fork");
    }
    if (pid == 0) {
	audit(className, methodName);
	/*NOTREACHED*/
    }
    if (waitpid(pid, &status, 0) < 0) {
	err(1, "waitpid");
    }
    if (WIFEXITED(status)) {
	if ((e = WEXITSTATUS(status)) < 0 || e > 2) {
	    errx(1, "unknown exit status: %d", e);
	}
	return e;
    }
    return (WIFSIGNALED(status) ? 3 : 4);
}
