/*
	$Id: Event.c 217 2006-05-29 09:39:55Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <err.h>
#include "Display.h"
#include "Auditor.h"
#include "Event.h"

#define EVENT_INVOCATION 0
#define EVENT_RECIPE 1

struct InvocationEvent { 
    struct Display *dpy;
    void (*dispatch)(union Event *);
    void (*free)(union Event *);
    unsigned char *class;
    unsigned char *method;
};

struct RecipeEvent { 
    struct Display *dpy;
    void (*dispatch)(union Event *);
    void (*free)(union Event *);
};

union Event {
    struct {
	struct Display *dpy;
	void (*dispatch)(union Event *);
	void (*free)(union Event *);
    } any;
    struct InvocationEvent invocation;
    struct RecipeEvent recipe;
};

static unsigned char *
getString(union InputStream in)
{
    unsigned char *data;
    ssize_t size;
    int s;

    if ((s = InputStream_readFully(in, &size, sizeof(size))) < 0
	|| (size = htonl(size)) < 0
	|| (data = malloc(size + 1)) == NULL
	|| (s = InputStream_readFully(in, data, size)) < 0) {
	errx(1, "InputStream_read() failed.");
    }
    data[size] = 0;
    return data;
}

static void
InvocationEvent_free(union Event *ev)
{
    free(ev->invocation.class);
    free(ev->invocation.method);
    free(ev);
}

static void
InvocationEvent_dispatch(union Event *ev)
{
    int data[2];
    int s;
    struct Display *dpy;

#ifdef DEBUG
    warnx("%s: class=%s, method=%s", __FUNCTION__,
	  ev->invocation.class, ev->invocation.method);
#endif
    s = Auditor_invoke(ev->invocation.class, ev->invocation.method);
#ifdef DEBUG
    warnx("%s: return %d", __FUNCTION__, s);
#endif

    dpy = Event_getDisplay(ev);
    data[0] = htonl(sizeof(s));
    data[1] = htonl(s);
    Display_sendData(dpy, data, sizeof(data));
}

static union Event *
InvocationEvent_create(struct Display *dpy, union InputStream in)
{
    struct InvocationEvent *ev;

    if ((ev = malloc(sizeof(struct InvocationEvent))) == NULL)
	errx(1, "malloc() failed.");
    ev->dpy = dpy;
    ev->dispatch = InvocationEvent_dispatch;
    ev->free = InvocationEvent_free;
    ev->class = getString(in);
    ev->method = getString(in);
    return (union Event *)ev;
}

static void
RecipeEvent_free(union Event *ev)
{
    free(ev);
}

static void
RecipeEvent_dispatch(union Event *ev)
{
    struct Display *dpy;

    dpy = Event_getDisplay(ev);
    Display_sendRecipe(dpy);
}

static union Event *
RecipeEvent_create(struct Display *dpy)
{
    struct RecipeEvent *ev;

    if ((ev = malloc(sizeof(struct RecipeEvent))) == NULL)
	errx(1, "malloc() failed.");
    ev->dpy = dpy;
    ev->dispatch = RecipeEvent_dispatch;
    ev->free = RecipeEvent_free;
    return (union Event *)ev;
}

union Event *
Event_create(struct Display *dpy, union InputStream in)
{
    int type;

    if (InputStream_readFully(in, &type, sizeof(type)) < 0)
	return NULL;
    type = htonl(type);
    switch (type) {
    default:
	errx(1, "unknown event: %d", type);
    case EVENT_INVOCATION:
	return InvocationEvent_create(dpy, in);
    case EVENT_RECIPE:
	return RecipeEvent_create(dpy);
    }
    /* NOTREACHED */
    return NULL;
}

void
Event_free(union Event *ev)
{
    if (ev == NULL)
	return;
    ev->any.free(ev);
}

void
Event_dispatch(union Event *ev)
{
    ev->any.dispatch(ev);
}

struct Display *
Event_getDisplay(union Event *ev)
{
    return ev->any.dpy;
}
