/*
	$Id: OutputStream.c 118 2006-03-08 20:11:48Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <err.h>
#include "OutputStream.h"
#include "impl/OutputStream.h"

#ifdef linux
#include <libio.h>

typedef struct {
    void *cookie;
    int (*readfn)(void *, char *, int);
    int (*writefn)(void *, const char *, int);
    /* seekfnηˤĤƤfunopen()ΥȤ򻲾ȤΤ */
    off_t (*seekfn)(void *, off_t, int);
    int (*closefn)(void *);
} BSD;

static int 
GNU_read(void *gnu, char *data, size_t size)
{
    BSD *bsd = (BSD *)gnu;

    return bsd->readfn(bsd->cookie, data, (int)size);
}

static int 
GNU_write(void *gnu, const char *data, size_t size)
{
    BSD *bsd = (BSD *)gnu;

    return bsd->writefn(bsd->cookie, data, (int)size);
}

static int 
GNU_seek(void *gnu, _IO_off64_t *pos, int whence)
{
    BSD *bsd = (BSD *)gnu;

    return ((*pos = bsd->seekfn(bsd->cookie, *pos, whence)) == -1);
}

static int 
GNU_close(void *gnu)
{
    BSD *bsd = (BSD *)gnu;
    int s;

    s = bsd->closefn(bsd->cookie);
    free(bsd);
    return s;
}

static const cookie_io_functions_t GNU_functions = {
    .read = GNU_read,
    .write = GNU_write,
    .seek = GNU_seek,
    .close = GNU_close};

static FILE *
funopen(const void *cookie,
	int (*readfn)(void *, char *, int),
	int (*writefn)(void *, const char *, int),
	/*
	  ʪfunopen(2)4ϼΤ褦Ƥ롣

	  fpos_t (*seekfn)(void *, fpos_t, int),

	  BSDǤfpos_t64ӥåʤΤǡseekfnθƤӽФ
	  lseek(2)ȥޥåƤȹͤƤ褤GNU libcǤ
	  fpos_tϹ¤ΤǤ뤿ᡢ嵭Ǥseekfnlseek(2)θ
	  ӽФϥޥåʤlseek(2)2ͤηǤ
	  off_tfpos_tߤѴǤʤΤǡΤޤޤǤϥߥ
	  졼ǤʤȤˤʤ롣

	  Τᡢfunopen(2)4뤳Ȥˤ롣
	  衢νϥꥸʥAPIŬѤ٤Ǥ롣Ʊ
	  readfn, writefnθƤӽФ⤽줾read(2), write(2)
	  ˥ޥå2char *Ǥʤvoid *3int
	  Ǥʤsize_tˤʤˤ褦˽٤
	*/
	off_t (*seekfn)(void *, off_t, int),
	int (*closefn)(void *))
{
    BSD *bsd;

    if ((bsd = (BSD *)malloc(sizeof(*bsd))) == NULL)
	return NULL;
    bsd->cookie = (void *)cookie;
    bsd->readfn = readfn;
    bsd->writefn = writefn;
    bsd->seekfn = seekfn;
    bsd->closefn = closefn;
    return fopencookie(bsd, "r+", GNU_functions);
}
#endif

void
impl_OutputStream_new(struct impl_OutputStream *impl,
		      ssize_t (*writefn)(struct impl_OutputStream *,
					 const void *, size_t),
		      int (*closefn)(struct impl_OutputStream *))
{
    impl->write = writefn;
    impl->close = closefn;
    impl->ancientStream = 0;
}

ssize_t
OutputStream_write(union OutputStream out, const void *data, size_t size)
{
    return out.impl->write(out.impl, data, size);
}

int
OutputStream_writeFully(union OutputStream out, const void *top, size_t size)
{
    int s;
    const char *data = top;

    while (size > 0 && (s = out.impl->write(out.impl, data, size)) > 0) {
        data += s;
        size -= s;
    }
    return (size > 0) ? -1 : 0;
}

int
OutputStream_close(union OutputStream out)
{
    if (out.impl->ancientStream) {
	warnx("%s: %d ancient stream(s) not closed.", __FUNCTION__,
	      out.impl->ancientStream);
    }
    return out.impl->close(out.impl);
}

union Instance {
    void *cookie;
    struct impl_OutputStream *impl;
} TRANSPARENT_UNION;

static int
writefn(union Instance this, const char *data, int size)
{
    return OutputStream_write(this.impl, data, size);
}

static int
closefn(union Instance this)
{
    --(this.impl->ancientStream);
    return 0;
}

FILE *
OutputStream_getAncientStream(union OutputStream out)
{
    ++(out.impl->ancientStream);
    return funopen(out.impl, NULL, writefn, NULL, closefn);
}

int
OutputStream_print(union OutputStream out, const char *format, ...)
{
    va_list ap;
    FILE *fp;
    int s;

    if ((fp = OutputStream_getAncientStream(out)) == NULL)
	return -1;
    va_start(ap, format);
    s = vfprintf(fp, format, ap);
    va_end(ap);
    (void)fclose(fp);
    return s;
}
