/*
	$Id: Socket.c 212 2006-05-27 21:06:01Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <err.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "Socket.h"

static int
newSocket(struct addrinfo *r)
{
    int fd;

    if ((fd = socket(r->ai_family, r->ai_socktype, r->ai_protocol)) < 0) {
	return -1;
    }
    if (connect(fd, r->ai_addr, r->ai_addrlen) < 0) {
	close(fd);
	return -1;
    }
    return fd;
}

int
Socket_open(const char *hostName, const char *portName)
{
    struct addrinfo h, *r, *res;
    int fd, s;

    bzero(&h, sizeof(h));
    h.ai_flags = AI_NUMERICHOST;
    h.ai_family = PF_UNSPEC;
    h.ai_socktype = SOCK_STREAM;
    if ((s = getaddrinfo(hostName, portName, &h, &res))) {
	h.ai_flags = AI_CANONNAME;
	s = getaddrinfo(hostName, portName, &h, &res);
    }
    if (s == EAI_SYSTEM)
	err(1, "%s", hostName);
    if (s)
	errx(1, "%s: %s", hostName, gai_strerror(s));
    if (res == NULL)
	errx(1, "%s: no addresses", hostName);
    for (r = res; r != NULL && (fd = newSocket(r)) < 0; r = r->ai_next)
	;
    if (r == NULL)
	err(1, "%s", hostName);
    freeaddrinfo(res);
    return fd;
}

void
Socket_setNoDelay(int fd)
{
    int opt;
    socklen_t optlen = sizeof(opt);

    if (getsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &opt, &optlen) < 0)
	err(1, "getsockopt");
    if (opt == 1)
	return;
    opt = 1;
    if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt)) < 0)
	err(1, "setsockopt");
}
