/*
	$Id: StackFrame.c 110 2006-03-04 21:05:48Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <dlfcn.h>
#include <stdio.h>
#include <bfd.h>
#include "A2LXlator.h"
#include "StackFrame.h"

#define FP_NEXT(x) ((StackFrame)((x)[0]))
#define FP_RETURN_IP(x) ((StackFrame)((x)[1]))
#define FP_IS_ROOT(x) (FP_NEXT(x) == 0 || FP_RETURN_IP(x) == 0)

extern int main(void);

static int
printSymbolLine(unsigned int vma)
{
    Dl_info info;
    const char *file, *function;
    int s, line;
    struct A2LXlator *t;
    void *mainBase;

    if (dladdr((void *)main, &info) == 0) {
	return 1;
    }
    mainBase = info.dli_fbase;

    if (dladdr((void *)vma, &info) == 0
	|| (t = A2LXlator_create(info.dli_fname)) == NULL) {
	return 1;
    }
    if (info.dli_fbase != mainBase) {
	vma -= (unsigned int)info.dli_fbase;
    }
    if ((s = A2LXlator_query(t, vma, &file, &function, &line)) == 0) {
	fprintf(stderr, "%s:%d: in funciton `%s`\n", file, line, function);
    }
    A2LXlator_free(t);
    return s;
}

static void
printWhereCalled(StackFrame fp)
{
    unsigned int vma = (unsigned int)FP_RETURN_IP(fp) - 1;

    if (printSymbolLine(vma)) {
	fprintf(stderr, "?:?: in funciton `??`\n");
    }
}

void
StackFrame_printStackTrace(StackFrame fp, int maxDepth)
{
    int k;

    for (k = 0; !FP_IS_ROOT(fp) && k < maxDepth; ++k, fp = FP_NEXT(fp)) {
	printWhereCalled(fp);
    }
    if (k == maxDepth) {
	fprintf(stderr, "...\n");
    }
}
